# AtlasCBS.py - AtlasCBS Ligand efficiency plugin script for StarDrop
# Version: 3f
# Copyright (C) Cele Abad-Zapatero, A. Cortes-Cabrera 2013
# Date: April 17, 2013. 
#
# This is python script file AtlasCBS.py
# incorporating the calculation of LEH (Hopkins et al. LE, DDT, 2004),
# plus the LEP (see below).
#
# This script enables the user to calculate the 
# LEIs variables (i.e. NBEI, NSEI and others)
# as a pair of variables to permit
# an Atlas-like representation of the content of the data set.
# The script derives Ki values from the pKi values indicated by the
# user from the data set. It could be modified to read also Ki's
# directly but for the time being, the presence of pKi's is assumed.
# Eight columns are added after the one containing the pKi assigned.
# NSEI,NBEI,BEI,SEI,nBEI,mBEI,LEH,LEP(DeltaG/NPOL)
# Details of the definitions can be found in the corresponding
# publications, except for the last one. This has been
# invented on March 22, 2013 as an extension/complement to LEH.
#   LEH (from Hopkins) (deltaG/Heavy) 
#   LEP (from C.A-Z)   (deltaG/Npol). 
# 
# This script derives from the original script provided
# by Ed Champness. The script has been written by
# Alvaro Cortes-Cabrera and Cele Abad-Zapatero,
# during the visit by C.A-Z to the UAH (sponsored by
# the Foundation Giner de los Rios) in the Depart.
# of Pharmacology, Facultad de Medicine. PI: Prof. F. Gago Badenas

# For advice or feedback, contact stardrop-support@optibrium.com

import app
import MoleculeUtils
import math
import PythonQt
import DataSetUtils

def getLE(dataset):
    ds = app.Dataset()
    ds.parse(dataset)

    # identify columns containing the variables needed
    # for the calculation of the various LEIs: MW, PSA for BEI, SEI respectively
    idTPSA = -1
    idMW = -1

    #   cols = DataSetUtils.getColumnNames(dataset)
    #   print cols.keys()
    for i in range(ds.columnSz()):
        if ds.columnName(i) == "MW":
            idMW = i
        elif ds.columnName(i) == "TPSA":
            idTPSA = i
    showcols = {}
    idcol = -1
    i = 0
    while i < ds.columnSz():
        type = ds.columnType(i)
        # find the structures
        if type == app.mol:
            idcol = i
            # remove columns of the wrong type
        if type == app.cont:
            showcols[ds.columnName(i)] = i
        i += 1
    if len(showcols) < 3:
        return 'No data available for calculation - pKi values required', False
    if idcol == -1:
        return 'No compound structures in data set available for calculation', False
    if idMW == -1:
        return 'No Molecular Weight values found in data set (column "MW")', False
    if idTPSA == -1:
        return 'No TPSA values found in data set (column "TPSA")', False
    else:
        dlg = PythonQt.ColumnPicker()
        dlg.setWindowTitle("AtlasCBS: Ligand Efficiency Indices")
        dlg.setLabel("Select column containing pKi values:")
        dlg.setColumns(showcols)
        res = dlg.run();
        if res == True:
            col = dlg.getColumn()
            # find the pKi values
            print('got column: ' + col)
            i = 0
            colindex = -1
            while i < ds.columnSz():
                if ds.columnName(i) == col:
                    colindex = i
                i += 1
            colindex = showcols.get(col)
            if colindex >= 0:
                return calculate(ds, colindex, idcol, idTPSA, idMW)
            else:
                return 'Inappropriate column selected', False


def calculate(ds, col, idcol, idTPSA, idMW):
    # calculate various ligand efficiency indices for each item in dataset
    # the first one is called le for historical reasons (original name given in the E.Champness script).
    # corresponds to nsei = pki/countpa (pki divided by polar atom count (N,O))

    # define the R (KiloCals/mol), T constants for the DeltaG calculation
    rconst = 0.00198
    tkelvin = 300.0
    # First efficiency calculated is NSEI
    efficiencies = []
    efficiencies_nbei = []
    efficiencies_bei = []
    efficiencies_sei = []
    efficiencies_nBEI = []
    efficiencies_mBEI = []
    efficiencies_leh = []
    efficiencies_lep = []

    for i in range(ds.rowSz()):
        smiles = ds.stringval(idcol, i)
        countpa = MoleculeUtils.countPolarAtoms(smiles)
        countha = MoleculeUtils.countHeavyAtoms(smiles)
        temp = app.Dataset.createInvalidEntry(app.cont)

        # get values of the different parameters needed to calculate the different
        # LEIs used in the AtlasCBS representation: MW, TPSA, LE, pki, etcc
        # and assign error estimates if possible (now hardwired).

        TPSA = ds.get(idTPSA, i)[1]
        erTPSA = ds.get(idTPSA, i)[2]
        MW = ds.get(idMW, i)[1]
        erMW = ds.get(idMW, i)[2]
        pki = ds.get(col, i)[1]
        pkierr = ds.get(col, i)[2]
        # compute ki from given pKi
        ki = math.pow(10, -pki)
        # Rather than just make up a ki err we an approximate it from the actual pki error
        kierr = math.log(10) * ki * pkierr

        # block for NSEI inherited from the old LE of E. Champness
        if countpa > 0:
            le = pki / countpa
            leerr = pkierr / countpa
        else:
            le = 0
            leerr = 0
        t1 = temp[0]
        t2 = ("True", "", "")
        t3 = (str(0), le, leerr)
        entry = (t1, t2, t3)
        efficiencies.append(entry)

        # block for NBEI
        if countha > 0:
            nbei = pki / countha
            nbeierr = pkierr / countha
        else:
            nbei = 0
            nbeierr = 0
        t1 = temp[0]
        t2 = ("True", "", "")
        t3 = (str(0), nbei, nbeierr)
        entry = (t1, t2, t3)
        efficiencies_nbei.append(entry)

        # block for BEI, using MW from dataset
        if MW > 0:
            bei = pki / (MW / 1000.)
            beierr = pkierr / (MW / 1000.)  # Assuming no error in MW
        else:
            bei = 0
            beierr = 0
        t1 = temp[0]
        t2 = ("True", "", "")
        t3 = (str(0), bei, beierr)
        entry = (t1, t2, t3)
        efficiencies_bei.append(entry)

        # block for SEI, using TPSA from dataset
        if TPSA > 0:
            sei = pki / (TPSA / 100.)
            seierr = pkierr / (TPSA / 100.)  # Assuming no error in TPSA
        else:
            sei = 0
            seierr = 0
        t1 = temp[0]
        t2 = ("True", "", "")
        t3 = (str(0), sei, seierr)
        entry = (t1, t2, t3)
        efficiencies_sei.append(entry)

        # block for nBEI, using countha, ki extracted from dataset above
        if countha > 0:
            ratio1 = ki / countha
            nBEI = -math.log10(ratio1)
            nBEIerr = pkierr
        else:
            nBEI = 0
            nBEIerr = 0
        t1 = temp[0]
        t2 = ("True", "", "")
        t3 = (str(0), nBEI, nBEIerr)
        entry = (t1, t2, t3)
        efficiencies_nBEI.append(entry)

        # block for mBEI, using MW calculated from model
        if MW > 0:
            ratio2 = ki / MW
            mBEI = - math.log10(ratio2)
            mBEIerr = pkierr
        else:
            mBEI = 0.0
            mBEIerr = 0.0
        t1 = temp[0]
        t2 = ("True", "", "")
        t3 = (str(0), mBEI, mBEIerr)
        entry = (t1, t2, t3)
        efficiencies_mBEI.append(entry)

        # block for LEH, using countha, ki extracted from dataset above
        # leh is the initial LE suggested by Hopkins et al. 2004 (DDT paper)
        if countha > 0:
            leh = rconst * tkelvin * math.log(ki)
            leh = -leh / countha
            leherr = rconst * tkelvin * kierr / (ki * countha)
        else:
            leh = 0
            leherr = 0
        t1 = temp[0]
        t2 = ("True", "", "")
        t3 = (str(0), leh, leherr)
        entry = (t1, t2, t3)
        efficiencies_leh.append(entry)

        # block for LEP, using countpa, ki extracted from dataset above
        # leh is the initial LE suggested by Hopkins et al. 2004 (DDT paper)
        # lep is a 'natural' extension for DeltaG/NPOL atoms (CAZ, unpublished).

        if countpa > 0:
            lep = rconst * tkelvin * math.log(ki)
            lep = -lep / countpa
            leperr = rconst * tkelvin * kierr / (ki * countpa)
        else:
            lep = 0
            leperr = 0
        t1 = temp[0]
        t2 = ("True", "", "")
        t3 = (str(0), lep, leherr)
        entry = (t1, t2, t3)
        efficiencies_lep.append(entry)

        # add new columns for NSEI, NBEI, BEI, SEI, up to LEH and LEP (this
        # is a 'natural' extension from LEH(NHA) of Hopkins etc...
        # The columns are added after the corresponding pKi column

        result_columns = []
        index = col + 1
    lecol = app.Dataset.meta('NSEI (' + ds.columnName(col) + ')', app.cont, ((0, 100), (0, 0)))
    ds.insertColumn(lecol, efficiencies, index)
    result_columns.append((ds.packColumn(index), index))
    index += 1

    lecol = app.Dataset.meta('NBEI (' + ds.columnName(col) + ')', app.cont, ((0, 100), (0, 0)))
    ds.insertColumn(lecol, efficiencies_nbei, index)
    result_columns.append((ds.packColumn(index), index))
    index += 1

    lecol = app.Dataset.meta('BEI (' + ds.columnName(col) + ')', app.cont, ((0, 100), (0, 0)))
    ds.insertColumn(lecol, efficiencies_bei, index)
    result_columns.append((ds.packColumn(index), index))
    index += 1

    lecol = app.Dataset.meta('SEI (' + ds.columnName(col) + ')', app.cont, ((0, 100), (0, 0)))
    ds.insertColumn(lecol, efficiencies_sei, index)
    result_columns.append((ds.packColumn(index), index))
    index += 1

    lecol = app.Dataset.meta('nBEI (' + ds.columnName(col) + ')', app.cont, ((0, 100), (0, 0)))
    ds.insertColumn(lecol, efficiencies_nBEI, index)
    result_columns.append((ds.packColumn(index), index))
    index += 1

    lecol = app.Dataset.meta('mBEI (' + ds.columnName(col) + ')', app.cont, ((0, 100), (0, 0)))
    ds.insertColumn(lecol, efficiencies_mBEI, index)
    result_columns.append((ds.packColumn(index), index))
    index += 1

    lecol = app.Dataset.meta('LEH (' + ds.columnName(col) + ')', app.cont, ((0, 100), (0, 0)))
    ds.insertColumn(lecol, efficiencies_leh, index)
    result_columns.append((ds.packColumn(index), index))
    index += 1

    lecol = app.Dataset.meta('LEP (' + ds.columnName(col) + ')', app.cont, ((0, 100), (0, 0)))
    ds.insertColumn(lecol, efficiencies_lep, index)
    result_columns.append((ds.packColumn(index), index))
    index += 1

    return result_columns, True


def calculateAtlasCBS(stardrop_appui):
    """ Select and execute models to ensure the existence of the necessary variables"""
    stardrop_appui.selectModel("MW")
    stardrop_appui.selectModel("TPSA")
    stardrop_appui.runModels()
    dataset = stardrop_appui.importDS("")
    result = getLE(dataset)
    if result[1] == False:
        if len(result[0]) > 0:
            stardrop_appui.show(result[0])
        else:
            stardrop_appui.show('Calculation failed')
    else:
        stardrop_appui.startMacro("")
        for (data, index) in result[0]:
            stardrop_appui.addColumnToDataSet('', index, data)
        stardrop_appui.stopMacro("")


def get_stardrop_definitions():
    definitions = [
        {
            'script_name': 'AtlasCBS: Ligand Efficiency Indices',
            'callback': calculateAtlasCBS
        }
    ]

    return definitions
